/*
 * $Id: HTMLCallbackHandler.java,v 1.2 2003/07/13 10:48:41 kenta Exp $
 *
 * Copyright (C) 2003 Kenta Cho. All rights reserved.
 */
package com.ecyrd.jspwiki.xmlrpc;

import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLEditorKit.ParserCallback;

/**
 * Convert HTML to Wiki text.
 */
public class HTMLCallbackHandler extends ParserCallback {
	StringBuffer buf;
	String currentURL, baseURL;
	private int ulNest, olNest;

	public HTMLCallbackHandler(
		StringBuffer buf,
		String currentURL,
		String baseURL) {
		this.buf = buf;
		this.currentURL = currentURL;
		this.baseURL = baseURL;
		ulNest = olNest = 0;
	}

	private void addCr() {
		if (buf.length() > 0 && buf.charAt(buf.length() - 1) == '\n') {
			return;
		}
		buf.append('\n');
	}

	public void handleStartTag(
		HTML.Tag tag,
		MutableAttributeSet set,
		int idx) {
		if (tag.equals(HTML.Tag.A)) {
			String href = (String) set.getAttribute(HTML.Attribute.HREF);
			if (href != null) {
				if (!href.startsWith("http://")) {
					if (!href.startsWith("/")) {
						href = currentURL + href;
					} else {
						href = baseURL + href;
					}
				}
				buf.append("[" + href + "]");
			}
		} else if (tag.equals(HTML.Tag.H1) || tag.equals(HTML.Tag.TITLE)) {
			addCr();
			buf.append("!!!");
		} else if (tag.equals(HTML.Tag.H2)) {
			addCr();
			buf.append("!!");
		} else if (tag.equals(HTML.Tag.H3)) {
			addCr();
			buf.append("!");
		} else if (tag.equals(HTML.Tag.UL)) {
			if (ulNest < 4)
				ulNest++;
		} else if (tag.equals(HTML.Tag.OL)) {
			if (olNest < 4)
				olNest++;
		} else if (tag.equals(HTML.Tag.LI)) {
			if (ulNest > 0) {
				addCr();
				for (int i = 0; i < ulNest; i++) {
					buf.append("*");
				}
			} else if (olNest > 0) {
				addCr();
				for (int i = 0; i < olNest; i++) {
					buf.append("#");
				}
			}
		} else if (tag.equals(HTML.Tag.P)) {
			addCr();
		}
	}

	public void handleEndTag(HTML.Tag tag, int idx) {
		if (tag.equals(HTML.Tag.UL)) {
			if (ulNest > 0)
				ulNest--;
		} else if (tag.equals(HTML.Tag.OL)) {
			if (olNest > 0)
				olNest--;
		} else if (
			tag.equals(HTML.Tag.P)
				|| tag.equals(HTML.Tag.TITLE)
				|| tag.equals(HTML.Tag.H1)
				|| tag.equals(HTML.Tag.H2)
				|| tag.equals(HTML.Tag.H3)) {
			addCr();
		}
	}

	public void handleSimpleTag(
		HTML.Tag tag,
		MutableAttributeSet set,
		int idx) {
		if (tag.equals(HTML.Tag.HR)) {
			addCr();
			buf.append("----");
			addCr();
		} else if (tag.equals(HTML.Tag.BR)) {
			addCr();
		}
	}

	private final int[] rmvChar = { '\n', '\r', '[', ']' };

	private String removeSpecificChars(String t) {
		StringBuffer buf = new StringBuffer();
		int i = 0;
		for (;;) {
			int mi = Integer.MAX_VALUE;
			for (int j = 0; j < rmvChar.length; j++) {
				int ci = t.indexOf(rmvChar[j], i);
				if (ci != -1 && ci < mi) {
					mi = ci;
				}
			}
			if (mi == Integer.MAX_VALUE) {
				buf.append(t.substring(i));
				break;
			}
			buf.append(t.substring(i, mi));
			i = mi + 1;
		}
		return buf.toString();
	}

	public void handleText(char[] aChar, int i) {
		String text = String.valueOf(aChar);
		buf.append(removeSpecificChars(text));
	}
}
