/*
 * $Id: RPCHandler4IWikiView.java,v 1.2 2003/07/13 10:48:41 kenta Exp $
 *
 * Copyright (C) 2003 Kenta Cho. All rights reserved.
 */
package com.ecyrd.jspwiki.xmlrpc;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;

import javax.swing.text.html.parser.ParserDelegator;

import org.apache.xmlrpc.XmlRpcException;

/**
 * RPCHandler for IWikiView.
 *  . Allows the 'wiki.putPage' call.
 *  . When the name starts with 'http://', get the Web page.
 *  . When the name starts with 'g@', get the Google result.
 */
public class RPCHandler4IWikiView extends RPCHandler {

	public boolean putPage(String page, byte[] tb) {
		try {
			String text = new String(tb, "UTF-8");
			m_engine.saveText(convertNCRtoString(page), text);
		} catch (Exception e) {
			return false;
		}
		return true;
	}

	private final int HTML_PAGE_MAX_LENGTH = 10240;

	private String getHTMLPage(String urlAdr) throws IOException {
		URL url = new URL(urlAdr);
		HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		conn.setRequestProperty("Referer", "");
		conn.setRequestProperty(
			"User-Agent",
			"Mozilla/4.0 (compatible;MSIE 5.5; Windows NT 5.0; H010818)");
		String type = conn.getContentType();
		if (!(type.startsWith("text/html") | type.startsWith("text/plain"))) {
			throw new IOException("Invalid content type");
		}
		String enc = "JISAutoDetect";
		if (type.indexOf("charset=UTF-8") != -1) {
			enc = "UTF-8";
		}
		BufferedReader reader =
			new BufferedReader(
				new InputStreamReader(conn.getInputStream(), enc));
		StringBuffer buf = new StringBuffer();
		String baseURL = urlAdr;
		if (baseURL.indexOf("/", 7) > 0) {
			baseURL = baseURL.substring(0, baseURL.indexOf("/", 7));
		}
		String currentURL = urlAdr;
		if (currentURL.lastIndexOf("/") > 6) {
			currentURL = currentURL.substring(0, currentURL.lastIndexOf("/"));
		}
		currentURL += "/";
		HTMLCallbackHandler handler =
			new HTMLCallbackHandler(buf, currentURL, baseURL);
		new ParserDelegator().parse(reader, handler, true);
		reader.close();
		String contents = buf.toString();
		if (contents.length() > HTML_PAGE_MAX_LENGTH) {
			contents = contents.substring(0, HTML_PAGE_MAX_LENGTH);
		}
		return contents;
	}

	private final String GOOGLE_RESULT_START_STR = "fBNg";
	private final String GOOGLE_REMOVE_URL = "[http://www.google.co.jp";

	private String adjustGoogleResult(String result) {
		String r = 
			result.substring(
				result.indexOf(GOOGLE_RESULT_START_STR)
					+ GOOGLE_RESULT_START_STR.length());
		StringBuffer buf = new StringBuffer();
		int idx = 0;
		for (;;) {
			int ri = r.indexOf(GOOGLE_REMOVE_URL, idx);
			if (ri == -1) {
				buf.append(r.substring(idx));
				break;
			}
			buf.append(r.substring(idx, ri));
			idx = r.indexOf("]", ri + GOOGLE_REMOVE_URL.length()) + 1;
			if (idx == -1) {
				break;
			}
		}
		return buf.toString();
	}

	private String getGooglePage(String keyword) throws IOException {
		String url =
			"http://www.google.co.jp/search?hl=ja&ie=UTF-8&oe=sjis&q="
				+ URLEncoder.encode(convertNCRtoString(keyword), "UTF-8");
		String result = getHTMLPage(url);
		return adjustGoogleResult(result);
	}

	private final String HTTP_PREFIX = "http://";
	private final String GOOGLE_PREFIX = "g@";

	public byte[] getPage(String pageName) throws XmlRpcException {
		if (pageName.startsWith(HTTP_PREFIX)) {
			try {
				String contents = getHTMLPage(pageName);
				return contents.getBytes("UTF-8");
			} catch (IOException e) {
				if (!pageName.endsWith("/")) {
					try {
						String contents = getHTMLPage(pageName + "/");
						return contents.getBytes("UTF-8");
					} catch (IOException e1) {
					}
				}
				throw new XmlRpcException(
					0,
					"get a web page failed. " + e.getMessage());
			}
		} else if (pageName.startsWith(GOOGLE_PREFIX)) {
			try {
				String contents =
					getGooglePage(pageName.substring(GOOGLE_PREFIX.length()));
				return contents.getBytes("UTF-8");
			} catch (IOException e) {
				throw new XmlRpcException(
					0,
					"get Google result failed. " + e.getMessage());
			}
		} else {
			return super.getPage(convertNCRtoString(pageName));
		}
	}

	private String convertNCRtoString(String is) {
		StringBuffer rs = new StringBuffer();
		boolean amp = false;
		for (int i = 0;;) {
			if (i >= is.length())
				break;
			char c = is.charAt(i);
			if (amp) {
				amp = false;
				if (c == '#') {
					int ei = is.indexOf(';', i);
					int n = Integer.parseInt(is.substring(i + 1, ei));
					char ch = (char) n;
					rs.append(ch);
					i = ei + 1;
					continue;
				} else {
					rs.append('&');
				}
			} else {
				if (c == '&') {
					amp = true;
					i++;
					continue;
				}
			}
			rs.append(c);
			i++;
		}
		return rs.toString();
	}
}
