/*
 * $Id: Iwv.java,v 1.4 2003/07/13 10:48:22 kenta Exp $
 *
 * Copyright (C) 2003 Kenta Cho. All rights reserved.
 */
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;

import javax.microedition.io.Connector;

import com.nttdocomo.ui.AnchorButton;
import com.nttdocomo.ui.Component;
import com.nttdocomo.ui.ComponentListener;
import com.nttdocomo.ui.Display;
import com.nttdocomo.ui.Font;
import com.nttdocomo.ui.Frame;
import com.nttdocomo.ui.Graphics;
import com.nttdocomo.ui.HTMLLayout;
import com.nttdocomo.ui.Label;
import com.nttdocomo.ui.IApplication;
import com.nttdocomo.ui.Panel;
import com.nttdocomo.ui.SoftKeyListener;
import com.nttdocomo.ui.TextBox;

/**
 * IWikiView main routine.
 */
public class Iwv
    extends IApplication
    implements ComponentListener, SoftKeyListener, Runnable {
    private WikiHandler wh;

    // Application state. 0- Wiki page 1- Menu 2- Edit 
    //                    3- Config 4- History 5- Open.
    private int st;

    // Fonts.
    private Font f;
    private Font fb;

    // Colors.
    private int clh1, clh2, clu, clo, cllr, clln;

    // Main wiki view panel.
    private Panel wp;
    private HTMLLayout whl;
    private String wtt;
    private Hashtable wac;

    // Menu panel.
    private Panel mp;
    private HTMLLayout mhl;
    private Hashtable mac;

    // Edit panel.
    private Panel ep;
    private TextBox[] etc;

    // Config panel.
    private Panel cp;
    private HTMLLayout chl;
    private TextBox cut, cwt, cr1t, cr2t;
    private String cus, cws, cr1s, cr2s;

    // History panel.
    private Panel hp;
    private HTMLLayout hhl;
    private Hashtable hac;

    // Open panel.
    private Panel op;
    private HTMLLayout ohl;
    private TextBox opn;

    // Network connection thread.
    //private Thread ct;
    // Network connection type. 0- getPage 1-putPage
    private int nct;
    // Network connection args.
    private String nca1, nca2;

    // Pages.
    private Hashtable pgs;
    // Page history.
    private Vector hst;

    // Current pages name.
    private String cn;

    // Application source URL.
    private String su;

    // Not configured(first time launch).
    private boolean nc;

    // Scratch pads sectors.
    public static final int SS = 1024;
    public static final int SN = 188;
    public static boolean[] so = new boolean[SN]; // Occupied or not.
    public static int sln; // # of unocuupied sectors.

    // Page records.
    private Vector rcd;

    /**
     * Load a page header form a scratch pad.
     * sector(n): dirty/name/contents bytes length/#sector/sector idx/idx/idx/...
     */
    private Page loadPageHeader(int n) throws IOException {
        Page p = new Page();
        p.ld = false;
        DataInputStream is =
            Connector.openDataInputStream("scratchpad:///0;pos=" + (n * SS));
        p.dt = is.readBoolean();
        p.nm = is.readUTF();
        int bl = is.readInt();
        p.ctbn = bl;
        p.sl = is.readInt();
        for (int i = 0; i < p.sl; i++) {
            p.sn[i] = is.readInt();
            so[p.sn[i]] = true;
            sln--;
        }
        is.close();
        return p;
    }

    /**
     * Load page contents from a scratch pad.
     * sector(idx): contents
     */
    private void loadPageContents(Page p) throws IOException {
        int bl = p.ctbn;
        if (bl <= 0) {
            p.ct = "";
            p.ld = true;
            return;
        }
        byte[] sb = new byte[bl];
        for (int i = 0; i < p.sl - 1; i++) {
            DataInputStream is =
                Connector.openDataInputStream(
                    "scratchpad:///0;pos=" + (p.sn[i + 1] * SS));
            if (bl >= SS) {
                is.read(sb, i * SS, SS);
                bl -= SS;
            } else {
                is.read(sb, i * SS, bl);
            }
            is.close();
        }
        p.ct = WikiHandler.getUTFString(sb);
        p.ld = true;
    }

    /**
     * Load the config from a scratch pad.
     * sector(0): XMLRPC URL/Start WikiName/'['replacer/']'replacer/
     *            Current WikiName/#recorded pages/page sector idx/idx/idx/...
     *            #history/history WikiName/h WN/h WN/... 
     */
    private void loadConfig() throws IOException {
        so[0] = true;
        sln = SN - 1;
        DataInputStream is = Connector.openDataInputStream("scratchpad:///0");
        cus = is.readUTF();
        cws = is.readUTF();
        if (cws == null || cws.length() <= 0) {
            is.close();
            throw new IOException("Invalid home wiki name");
        }
        cr1s = is.readUTF();
        cr2s = is.readUTF();
        cn = is.readUTF();
        int rs = is.readInt();
        int[] rci = new int[rs];
        for (int i = 0; i < rs; i++) {
            rci[i] = is.readInt();
        }
        int hn = is.readInt();
        for (int i = 0; i < hn; i++) {
            hst.addElement(is.readUTF());
        }
        is.close();

        for (int i = 0; i < rs; i++) {
            Page p = loadPageHeader(rci[i]);
            rcd.addElement(p);
            pgs.put(p.nm, p);
        }
    }

    /**
     * Touched page is moved to the top of the record.
     */
    private void touchPage(Page p) {
        rcd.removeElement(p);
        rcd.insertElementAt(p, 0);
    }

    private void removePage(Page p) {
        for (int i = 0; i < p.sl; i++) {
            so[p.sn[i]] = false;
        }
        sln += p.sl;
        rcd.removeElement(p);
        pgs.remove(p.nm);
        try {
            saveConfig();
        } catch (IOException e) {}
    }

    private int getUnoccupiedSectorIndex() {
        int i;
        // Sector(0) is used for config.
        for (i = 1; i < SN; i++) {
            if (!so[i])
                break;
        }
        return i;
    }

    /**
     * Save config to sector(0).
     */
    private void saveConfig() throws IOException {
        DataOutputStream os = Connector.openDataOutputStream("scratchpad:///0");
        os.writeUTF(cus);
        os.writeUTF(cws);
        os.writeUTF(cr1s);
        os.writeUTF(cr2s);
        os.writeUTF(cn);
        int rs = rcd.size();
        os.writeInt(rs);
        for (int i = 0; i < rs; i++) {
            Page p = (Page) rcd.elementAt(i);
            os.writeInt(p.sn[0]);
        }
        int hn = hst.size(), i;
        if (hn > 16) {
            i = hn - 16;
        } else {
            i = 0;
        }
        os.writeInt(hn - i);
        for (; i < hn; i++) {
            os.writeUTF((String) hst.elementAt(i));
        }
        os.close();
    }

    /**
     * Save a sector that stores the page header.
     * @param p page
     */
    private void savePageHeader(Page p) throws IOException {
        int si = p.sn[0];
        DataOutputStream os =
            Connector.openDataOutputStream("scratchpad:///0;pos=" + SS * si);
        os.writeBoolean(p.dt);
        os.writeUTF(p.nm);
        os.writeInt(p.ctbn);
        os.writeInt(p.sl);
        for (int i = 0; i < p.sl; i++) {
            os.writeInt(p.sn[i]);
        }
        os.close();
    }

    /**
     * Save a sector that stores the page contents fragment.
     * @param ci contents fragment index
     * @param fl contents fragment length
     * @param p page
     */
    private void savePageContents(int ci, int fl, Page p) throws IOException {
        int si = p.sn[ci + 1];
        DataOutputStream os =
            Connector.openDataOutputStream("scratchpad:///0;pos=" + SS * si);
        os.write(p.ctb, ci * SS, fl);
        os.close();
    }

    private void savePage(Page p) throws IOException {
        if (sln < p.sl) {
            int i = rcd.size();
            i--;
            for (; i >= 0; i--) {
                Page rp = (Page) rcd.elementAt(i);
                if (!rp.dt) {
                    removePage(rp);
                    if (sln >= p.sl) {
                        break;
                    }
                }
            }
            i = rcd.size();
            i--;
            for (; i >= 0; i--) {
                Page rp = (Page) rcd.elementAt(i);
                removePage(rp);
                if (sln >= p.sl) {
                    break;
                }
            }
        }
        for (int i = 0; i < p.sl; i++) {
            int n = getUnoccupiedSectorIndex();
            so[n] = true;
            sln--;
            p.sn[i] = n;
        }
        for (int i = 0; i < p.sl; i++) {
            if (i == 0) {
                savePageHeader(p);
            } else {
                if (i == p.sl - 1) {
                    savePageContents(i - 1, p.ctbn % SS, p);
                } else {
                    savePageContents(i - 1, SS, p);
                }
            }
        }
        pgs.put(p.nm, p);
        rcd.insertElementAt(p, 0);
        saveConfig();
    }

    /**
     * Initialize and create UI panels.
     */
    private void init() {
        pgs = new Hashtable();
        rcd = new Vector();
        hst = new Vector();
        try {
            loadConfig();
            nc = false;
        } catch (IOException e) {
            cus = ":8080/JSPWiki/RPC2/";
            cws = "Main";
            cr1s = "@,";
            cr2s = ",@";
            nc = true;
        }

        wh = new WikiHandler();
        su = IApplication.getCurrentApp().getSourceURL();
        if (su.indexOf("/", 7) > 0) {
            su = su.substring(0, su.indexOf("/", 7));
        }
        if (su.indexOf(":", 7) > 0) {
            su = su.substring(0, su.indexOf(":", 7));
        }
        wh.setURL(su + cus);

        f = Font.getFont(Font.FACE_SYSTEM | Font.SIZE_TINY | Font.STYLE_PLAIN);
        fb = Font.getFont(Font.FACE_SYSTEM | Font.SIZE_TINY | Font.STYLE_BOLD);

        clh1 = Graphics.getColorOfRGB(0x80, 0x10, 0x10);
        clh2 = Graphics.getColorOfRGB(0x70, 0x40, 0x40);
        clu = Graphics.getColorOfRGB(0x20, 0x80, 0x20);
        clo = Graphics.getColorOfRGB(0x20, 0x80, 0x40);
        cllr = Graphics.getColorOfRGB(0x20, 0x30, 0x40);
        clln = Graphics.getColorOfRGB(0x20, 0x50, 0xa0);

        startAddLines();
        endAddLines();

        mp = new Panel();
        mp.setTitle("j[");
        mp.setComponentListener(this);
        mp.setSoftKeyListener(this);
        mp.setSoftLabel(Frame.SOFT_KEY_1, "߂");
        mhl = new HTMLLayout();
        mp.setLayoutManager(mhl);
        mhl.begin(HTMLLayout.LEFT);
        mac = new Hashtable();
        AnchorButton ab = new AnchorButton("ҏW");
        mac.put(ab, "Edit");
        mp.add(ab);
        mhl.br();
        ab = new AnchorButton("y[WM");
        mac.put(ab, "Send");
        mp.add(ab);
        mhl.br();
        ab = new AnchorButton("y[Wēǂݍ");
        mac.put(ab, "Reload");
        mp.add(ab);
        mhl.br();
        ab = new AnchorButton("");
        mac.put(ab, "History");
        mp.add(ab);
        mhl.br();
        ab = new AnchorButton("y[WJ");
        mac.put(ab, "Open");
        mp.add(ab);
        mhl.br();
        ab = new AnchorButton("z[");
        mac.put(ab, "Home");
        mp.add(ab);
        mhl.br();
        ab = new AnchorButton("ݒ");
        mac.put(ab, "Config");
        mp.add(ab);
        mhl.br();
        ab = new AnchorButton("I");
        mac.put(ab, "Exit");
        mp.add(ab);
        mhl.end();

        cp = new Panel();
        cp.setSoftKeyListener(this);
        cp.setSoftLabel(Frame.SOFT_KEY_1, "ݒ");
        cp.setSoftLabel(Frame.SOFT_KEY_2, "~");
        cp.setTitle("ݒ");
        chl = new HTMLLayout();
        cp.setLayoutManager(chl);
        chl.begin(HTMLLayout.LEFT);
        cp.add(new Label("Wiki XML-RPC URL"));
        chl.br();
        cp.add(new Label(su));
        cut = new TextBox(cus, 99, 1, TextBox.DISPLAY_ANY);
        cp.add(cut);
        cp.add(new Label("z[ WikiName"));
        cwt = new TextBox(cws, 99, 1, TextBox.DISPLAY_ANY);
        cp.add(cwt);
        cp.add(new Label("'[' ֕"));
        cr1t = new TextBox(cr1s, 99, 1, TextBox.DISPLAY_ANY);
        cp.add(cr1t);
        cp.add(new Label("']' ֕"));
        cr2t = new TextBox(cr2s, 99, 1, TextBox.DISPLAY_ANY);
        cp.add(cr2t);
        chl.end();

        op = new Panel();
        op.setSoftKeyListener(this);
        op.setSoftLabel(Frame.SOFT_KEY_1, "J");
        op.setSoftLabel(Frame.SOFT_KEY_2, "~");
        op.setTitle("y[WJ");
        ohl = new HTMLLayout();
        op.setLayoutManager(ohl);
        ohl.begin(HTMLLayout.LEFT);
        op.add(new Label("WikiName"));
        opn = new TextBox("", 99, 1, TextBox.DISPLAY_ANY);
        op.add(opn);
        ohl.end();
    }

    /**
     * Change the front UI panel.
     * @param s 0- Wiki page. 1- Menu. 2- Edit 3- Config 4- History.
     */
    private void changeState(int s) {
        st = s;
        switch (s) {
            case 0 :
                Display.setCurrent(wp);
                break;
            case 1 :
                Display.setCurrent(mp);
                break;
            case 2 :
                Display.setCurrent(ep);
                break;
            case 3 :
                Display.setCurrent(cp);
                break;
            case 4 :
                Display.setCurrent(hp);
                break;
            case 5 :
                Display.setCurrent(op);
                break;
        }
    }

    /*
     *  Create the main wiki view panel.
     */

    private void setWikiPanelSoftLabel() {
        if (hst.size() > 1) {
            wp.setSoftLabel(Frame.SOFT_KEY_1, "߂");
        } else {
            wp.setSoftLabel(Frame.SOFT_KEY_1, null);
        }
    }

    private void startAddLines() {
        wp = new Panel();
        wp.setComponentListener(this);
        wp.setSoftKeyListener(this);
        whl = new HTMLLayout();
        wp.setLayoutManager(whl);
        whl.begin(HTMLLayout.LEFT);
        setWikiPanelSoftLabel();
        wp.setSoftLabel(Frame.SOFT_KEY_2, "ƭ");
        wac = new Hashtable();
    }

    private void endAddLines() {
        whl.end();
    }

    private void startAddHistories() {
        hp = new Panel();
        hp.setTitle("");
        hp.setComponentListener(this);
        hp.setSoftKeyListener(this);
        hhl = new HTMLLayout();
        hp.setLayoutManager(hhl);
        hhl.begin(HTMLLayout.LEFT);
        hp.setSoftLabel(Frame.SOFT_KEY_1, "߂");
        hac = new Hashtable();
    }

    private void endAddHistories() {
        hhl.end();
    }

    private void addText(String s) {
        AnchorButton a = new AnchorButton();
        a.setFont(f);
        a.setEnabled(false);
        if (s.startsWith("!!")) {
            s = s.substring(2);
            a.setFont(fb);
            a.setForeground(clh1);
        } else if (s.startsWith("!")) {
            s = s.substring(1);
            a.setFont(fb);
            a.setForeground(clh2);
        } else if (s.startsWith("*")) {
            a.setForeground(clu);
        } else if (s.startsWith("#")) {
            a.setForeground(clo);
        }
        a.setText(s);
        wp.add(a);
    }

	/**
	 * Change a first character of WikiName to the upper case.
	 */
    private String convertToWikiName(String l) {
        char tc = l.charAt(0);
        if (!l.startsWith("http://")
            && !l.startsWith("g@")
            && tc >= 'a'
            && tc <= 'z') {
            tc += 'A' - 'a';
            StringBuffer sb = new StringBuffer();
            sb.append(tc);
            sb.append(l.substring(1));
            return sb.toString();
        } else {
            return l;
        }
    }

    private void addAnchor(String s) {
        String t, l;
        int i = s.indexOf('|');
        if (i != -1) {
            t = s.substring(0, i);
            l = s.substring(i + 1);
        } else {
            t = l = s;
        }
        AnchorButton a = new AnchorButton();
        a.setFont(f);
        a.setEnabled(true);
        a.setText(t);
        l = convertToWikiName(l);
        if (pgs.containsKey(l)) {
            a.setForeground(cllr);
        } else {
            a.setForeground(clln);
        }
        wac.put(a, l);
        wp.add(a);
    }

    private void addHistory(String s) {
        AnchorButton a = new AnchorButton();
        a.setFont(f);
        a.setEnabled(true);
        a.setText(s);
        hac.put(a, s);
        hp.add(a);
        hhl.br();
    }

    /**
     * Add a text line to the panel and replace 
     * the WikiName with the AnchorButton.
     * (for JSPWiki)
     * @param s Wiki text
     */
    private void addLine(String s) {
        s.trim();
        if (s.length() <= 0)
            return;
        if (s.charAt(0) == '\n')
            return;
        if (s.charAt(0) == '\r')
            return;
        int i = 0;
        for (;;) {
            int be = s.indexOf(']', i);
            if (be == -1) {
                addText(s.substring(i));
                break;
            }
            int bs = s.indexOf('[', i);
            if (bs != -1 && bs < be) {
                addText(s.substring(i, bs));
                String as = s.substring(bs + 1, be);
                as.trim();
                if (as.length() > 0) {
                    if (as.charAt(0) == '[') {
                        addText(as + ']');
                    } else {
                        addAnchor(as);
                    }
                }
            } else {
                addText(s.substring(i, be + 1));
            }
            i = be + 1;
        }
        whl.br();
    }

    private void addPageString(String s) {
        int i = 0;
        for (;;) {
            int ci = s.indexOf('\n', i);
            if (ci == -1) {
                addLine(s.substring(i));
                return;
            }
            addLine(s.substring(i, ci));
            i = ci + 1;
        }
    }

    private void drawPage(String ct) {
        startAddLines();
        whl.br();
        addPageString(ct);
        endAddLines();
    }

    private void wikiPanelSetTitle(String t) {
        wtt = t;
        wp.setTitle(t);
    }

    private void showPage(String nm) {
        if (hst.size() == 0 || !nm.equals((String) hst.lastElement())) {
            hst.addElement(nm);
        }
        Page p = (Page) pgs.get(nm);
        drawPage(p.ct);
        if (p.dt) {
            wikiPanelSetTitle("* " + nm);
        } else {
            wikiPanelSetTitle(nm);
        }
        cn = nm;
        setWikiPanelSoftLabel();
        changeState(0);
    }

    private void gotoPage(String nm) {
        Page p = (Page) pgs.get(nm);
        if (p == null) {
            startGetPage(nm);
        } else {
            if (!p.ld) {
                // Page contents are not loaded.
                try {
                    loadPageContents(p);
                } catch (IOException e1) {}
            }
            touchPage(p);
            showPage(nm);
        }
        try {
            saveConfig();
        } catch (IOException e) {}
    }

    /**
     * Create the history panel.
     */
    private void gotoHistoryPage() {
        int hs = rcd.size();
        startAddHistories();
        for (int i = 0; i < hs; i++) {
            Page p = (Page) rcd.elementAt(i);
            addHistory(p.nm);
        }
        endAddHistories();
        changeState(4);
    }

    /**
     * Create the edit panel.
     */
    private void setEditPanel(String c) {
        ep = new Panel();
        ep.setTitle("ҏW - " + cn);
        ep.setSoftKeyListener(this);
        ep.setSoftLabel(Frame.SOFT_KEY_1, "");
        ep.setSoftLabel(Frame.SOFT_KEY_2, "~");

        int tn = (int) (c.length() / 200) + 1;
        etc = new TextBox[tn];
        for (int i = 0; i < tn; i++) {
            String ts;
            if (i == tn - 1) {
                ts = c.substring(i * 200);
            } else {
                ts = c.substring(i * 200, i * 200 + 200);
            }
            etc[i] = new TextBox(ts, 99, 4, TextBox.DISPLAY_ANY);
            ep.add(etc[i]);
            etc[i].setInputMode(TextBox.KANA);
        }
    }

    /**
     * An AnchorButton is pressed.
     */
    public void componentAction(Component c, int type, int prm) {
        if (c instanceof AnchorButton) {
            switch (st) {
                case 0 :
                    gotoPage((String) wac.get(c));
                    break;
                case 4 :
                    gotoPage((String) hac.get(c));
                    break;
                case 1 :
                    String lb = (String) mac.get(c);
                    if (lb.equals("Edit")) {
                        Page pg = (Page) pgs.get(cn);
                        setEditPanel(pg.ct);
                        changeState(2);
                        etc[0].requestFocus();
                    } else if (lb.equals("Send")) {
                        Page pg = (Page) pgs.get(cn);
                        changeState(0);
                        startPutPage(pg.nm, pg.ct);
                    } else if (lb.equals("Reload")) {
                        if (pgs.containsKey(cn)) {
                            Page p = (Page) pgs.get(cn);
                            removePage(p);
                        }
                        gotoPage(cn);
                    } else if (lb.equals("History")) {
                        gotoHistoryPage();
                    } else if (lb.equals("Open")) {
                        changeState(5);
                    } else if (lb.equals("Home")) {
                        gotoPage(cws);
                    } else if (lb.equals("Config")) {
                        changeState(3);
                    } else if (lb.equals("Exit")) {
                        exit();
                    }
                    break;
            }
        }
    }

    /*
     * Handle the connection to the wiki server.
     */

    private void startGetPage(String n) {
        nct = 0;
        nca1 = n;
        wp.setTitle("M");
        changeState(0);
        //ct = new Thread(this);
        //ct.start();
        run();
    }

    private void startPutPage(String n, String c) {
        nct = 1;
        nca1 = n;
        nca2 = c;
        wp.setTitle("M");
        changeState(0);
        //ct = new Thread(this);
        //ct.start();
        run();
    }

    /**
     * Thread to connect to the Wiki server.
     */
    public void run() {
        String ct = null;
        try {
            switch (nct) {
                case 0 :
                    ct = wh.getPage(nca1);
                    break;
                case 1 :
                    wh.putPage(nca1, nca2);
                    Page p = (Page) pgs.get(cn);
                    p.dt = false;
                    wikiPanelSetTitle(cn);
                    break;
            }
        } catch (NoSuchElementException e) {
            switch (nct) {
                case 0 :
                    if (nca1.startsWith("http://")) {
                        wp.setTitle("Ms: " + wtt);
                        break;
                    }
                    // New wiki page should be created.
                    ct = "";
                    break;
            }
        } catch (IOException e) {
            switch (nct) {
                case 0 :
                    wp.setTitle("Ms: " + wtt);
                    break;
                case 1 :
                    wp.setTitle("Ms: " + wtt);
                    break;
            }
        } finally {
            switch (nct) {
                case 0 :
                    if (ct == null)
                        break;
                    try {
                        Page p = new Page(nca1, ct);
                        p.dt = false;
                        savePage(p);
                        showPage(nca1);
                    } catch (IOException e1) {
                        wp.setTitle("y[W\s: " + wtt);
                    }
                    break;
                case 1 :
                    Page p = (Page) pgs.get(cn);
                    try {
                        savePageHeader(p);
                    } catch (IOException e2) {}
                    break;
            }
        }
    }

    /**
     * Handle soft keys in Wiki view Panel.
     * @param k key state
     */
    public void handleWpKey(int k) {
        if (k == Frame.SOFT_KEY_1) {
            if (hst.size() > 1) {
                int i = hst.size() - 2;
                String nm = (String) hst.elementAt(i);
                gotoPage(nm);
                hst.removeElementAt(i + 1);
                hst.removeElementAt(i + 1);
                setWikiPanelSoftLabel();
            }
        } else if (k == Frame.SOFT_KEY_2) {
            changeState(1);
        }
    }

    /**
     * Handle soft keys in menu panel.
     * @param k
     */
    public void handleMpKey(int k) {
        if (k == Frame.SOFT_KEY_1) {
            changeState(0);
        }
    }

    /**
     * Replace the specific string to the bracket
     * to save time to input the WikiName with the cellular phone.
     * (for JSPWiki)
     */
    private String replaceToWikiNameBracket(String s) {
        StringBuffer sb = new StringBuffer();
        String ss;
        if (cr1s.length() > 0) {
            for (int i = 0;;) {
                int bsi = s.indexOf(cr1s, i);
                if (bsi == -1) {
                    sb.append(s.substring(i));
                    break;
                }
                sb.append(s.substring(i, bsi));
                sb.append("[");
                i = bsi + cr1s.length();
            }
            ss = sb.toString();
        } else {
            ss = s;
        }
        if (cr2s.length() > 0) {
            sb = new StringBuffer();
            for (int i = 0;;) {
                int bsi = ss.indexOf(cr2s, i);
                if (bsi == -1) {
                    sb.append(ss.substring(i));
                    break;
                }
                sb.append(ss.substring(i, bsi));
                sb.append("]");
                i = bsi + cr2s.length();
            }
            return sb.toString();
        } else {
            return ss;
        }
    }

    /**
     * Handle soft keys in edit panel.
     */
    public void handleEpKey(int k) {
        if (k == Frame.SOFT_KEY_1) {
            Page p = (Page) pgs.get(cn);
            StringBuffer sb = new StringBuffer();
            for (int i = 0; i < etc.length; i++) {
                sb.append(etc[i].getText());
            }
            String ec = sb.toString();
            if (!p.ct.equals(ec)) {
                removePage(p);
                try {
                    p.setContents(replaceToWikiNameBracket(ec));
                    p.dt = true;
                    savePage(p);
                } catch (IOException e) {
                    wp.setTitle("Z[us: " + wtt);
                }
            }
            changeState(0);
            gotoPage(p.nm);
        } else if (k == Frame.SOFT_KEY_2) {
            changeState(0);
        }
    }

    /**
     * Handle soft keys in config panel.
     */
    public void handleCpKey(int k) {
        if (k == Frame.SOFT_KEY_1) {
            cus = cut.getText();
            cws = cwt.getText();
            cr1s = cr1t.getText();
            cr2s = cr2t.getText();
            wh.setURL(su + cus);
            if (cn == null)
                cn = cws;
            try {
                saveConfig();
            } catch (IOException e) {
                wp.setTitle("Z[us: " + wtt);
            }
            if (nc) {
                gotoPage(cws);
            }
            changeState(0);
        } else if (k == Frame.SOFT_KEY_2) {
            changeState(0);
        }
    }

    /**
     * Handle soft keys in history panel.
     */
    public void handleHpKey(int k) {
        if (k == Frame.SOFT_KEY_1) {
            changeState(0);
        }
    }

    /**
     * Handle soft keys in oepn panel.
     */
    public void handleOpKey(int k) {
        if (k == Frame.SOFT_KEY_1) {
        	String nm = opn.getText();
        	if (nm == null || nm.length() <= 0) {
        		changeState(0);
			} else {
	            gotoPage(convertToWikiName(nm));
			}
        } else if (k == Frame.SOFT_KEY_2) {
            changeState(0);
        }
    }

    public void softKeyPressed(int k) {}

    public void softKeyReleased(int k) {
        switch (st) {
            case 0 :
                handleWpKey(k);
                break;
            case 1 :
                handleMpKey(k);
                break;
            case 2 :
                handleEpKey(k);
                break;
            case 3 :
                handleCpKey(k);
                break;
            case 4 :
                handleHpKey(k);
                break;
            case 5 :
                handleOpKey(k);
                break;
        }
    }

    public void start() {
        init();
        if (nc) {
            // Open the config panel at the first launch time.
            changeState(3);
        } else {
            gotoPage(cn);
        }
    }

    private void exit() {
        terminate();
    }
}
